import fetch, {
  Request,
  Response,
} from 'node-fetch'
import { createLog } from './log';
import type { Logger } from 'winston';
import dotenv from 'dotenv';
import fs from 'fs/promises';
import { spawn } from 'child_process';
import type { ChildProcess } from 'child_process';

dotenv.config();

const log : Logger = createLog('client');

class Client {
	private maxJobs : number = 1;
	private url : string = 'http://localhost:3333';
	private jobs : string[] = [];

	constructor () {
		if (typeof process.env.YOLO_WEB_URL !== 'undefined') {
			this.url = process.env.YOLO_WEB_URL;
		}
		this.cron();
	}

	public cron () {
		this.job();
		setInterval(this.job.bind(this), 60 * 1000);
	}

	private job () {
		if (this.jobs.length >= this.maxJobs) {
			return;
		}
		this.poll();
	}

	private async poll () {
		const options : any = {
			method : 'GET'
		};
		const url : string = `${this.url}/job`;
		let res : Response;
		let json : any;

		try {
			log.info('Checking for new jobs...');
			res = await fetch(url, options);
		} catch (err) {
			log.error(`ERROR getting ${url}`);
			log.error(err);
			process.exit(1);
			return;
		}
		
		try {
			json = await res.json();
		} catch (err) {
			log.error('ERROR parsing JSON');
			log.error(err);
		}
		
		if (json === null || json.length === 0) {
			log.info('No jobs found.');
			return;
		}

		await this.claim(json[0]);
	}

	private async claim (id : string) {
		const options : any = {
			method : 'POST'
		};
		const url : string = `${this.url}/job/claim/${id}`;
		let res : Response;
		let json : any;

		try {
			log.info(`Claiming job ${id}...`);
			res = await fetch(url, options);
		} catch (err) {
			log.error(`ERROR posting to ${url}`);
			log.error(err);
			process.exit(1);
			return;
		}
		
		try {
			json = await res.json();
			log.info(json);
		} catch (err) {
			log.error('ERROR parsing JSON');
			log.error(err);
			process.exit(2);
			return; 
		}

		if (this.jobs.indexOf(id) === -1) {
			this.jobs.push(id);
			log.info(`Claimed job ${id}`);
			this.train(json);
		}
	}

	private async train (job : any) {
		const cmd : string = 'bash';
		const args : string[] = ['./scripts/train.sh', JSON.stringify(job)];
		const child : ChildProcess = spawn(cmd, args);
		
		child.stdout.on('data', (data : Buffer) => {
			const line : string = data.toString().trim();
			if (line === '') return;
			log.info(line);
		});

		child.stderr.on('data', (data : Buffer) => {
			const line : string = data.toString().trim();
			if (line === '') return;
			log.info(line);
    	});

		/*child.on('close', (code : number) => {
			log.info(`child process close all stdio with code ${code}`);
		});*/

		child.on('exit', (code : number) => {
			log.info(`Train job completed with code ${code}`);
			if (code !== 0) {
				this.fail(job.id, `Train script exited with code ${code}`)
			}
		});

		child.on('error', (err : Error) => {
			log.error('train job errored out');
			log.error(err);
			this.fail(job.id, err.toString());
		});
	}

	private async fail (id : string, reason : string) {
		const meta : any = {
			meta : reason
		};
		const options : any = {
			method : 'POST',
			headers: {
				'Accept': 'application/json',
				'Content-Type': 'application/json'
			},
			body: JSON.stringify(meta)
		};
		const url : string = `${this.url}/job/fail/${id}`;
		let res : Response;
		let json : any;

		this.removeJob(id);

		try {
			log.info(`Failed job ${id}...`);
			res = await fetch(url, options);
		} catch (err) {
			log.error(`ERROR posting to ${url}`);
			log.error(err);
			process.exit(1);
			return;
		}
		
		try {
			json = await res.json();
			log.info(json);
		} catch (err) {
			log.error('ERROR parsing JSON');
			log.error(err);
			process.exit(2);
			return; 
		}

		log.info(json);
	}

	private removeJob (id : string) {
		const index : number = this.jobs.indexOf(id);
		if (index !== -1) {
			this.jobs.splice(index, 1);
		}
	}
}

new Client()